@@ -29,6 +29,17 @@ xpath_registry = {}
 # EXSLT regular expression namespace URI
 REGEXP_NAMESPACE = "http://exslt.org/regular-expressions"
 
+class BadStatusError(AssertionError):
+    """
+    Raised when a non-success HTTP response is found (eg 404 or 500)
+    """
+
+class NotARedirectError(AssertionError):
+    """
+    Raised when an attempt is made to call follow() on a non-redirected
+    response
+    """
+
 class XPathMultiMethod(object):
     """
     A callable object that has different implementations selected by XPath
@@ -791,7 +802,7 @@ class TestAgent(object):
 
         response = self.response_class.from_wsgi(self.app, environ, self.start_response)
         if check_status and response.status[0] not in '23':
-            raise AssertionError("%s %r returned HTTP status %r" % (
+            raise BadStatusError("%s %r returned HTTP status %r" % (
                 environ['REQUEST_METHOD'],
                 path,
                 response.status
@@ -1004,6 +1015,56 @@ class TestAgent(object):
             check_status=check_status,
         )
 
+    def reload(self, follow=True, check_status=True):
+        """
+        Reload the current page, if necessary re-posting any data.
+
+        Form fields that have been filled in on the loaded page, they will be
+        refilled on the reloaded page, provided that the reloaded page has
+        exactly the same fields present in the same order.
+        """
+        env = self.request.environ.copy()
+        del env['pesto.request']
+        env['wsgi.input'].seek(0)
+        agent = self._request(env, follow=follow, check_status=check_status)
+
+        for src, dst in zip(
+            self.find('//input|//textarea|//option'),
+            agent.find('//input|//textarea|//option')
+        ):
+            if not all([
+                src.tag == dst.tag,
+                src.attrib.get('name') == dst.attrib.get('name'),
+                src.attrib.get('type') == dst.attrib.get('type'),
+            ]):
+                break
+
+            if (
+                src.tag == 'input'
+                and src.attrib.get('type') in ('radio', 'checkbox')
+            ):
+                if 'checked' in src.attrib:
+                    dst.attrib['checked'] = src.attrib['checked']
+                elif 'checked' in dst.attrib:
+                    del dst.attrib['checked']
+
+            elif src.tag == 'option':
+                if 'selected' in src.attrib:
+                    dst.attrib['selected'] = src.attrib['selected']
+                elif 'selected' in dst.attrib:
+                    del dst.attrib['selected']
+
+            elif src.tag == 'textarea':
+                dst.element.text = src.element.text
+
+            else:
+                if 'value' in src.attrib:
+                    dst.attrib['value'] = src.attrib['value']
+                elif 'value' in dst.attrib:
+                    del dst.attrib['value']
+
+        return agent
+
     def start_response(self, status, headers, exc_info=None):
         """
         No-op implementation.
@@ -1078,7 +1139,7 @@ class TestAgent(object):
         target. No entry is recorded in the agent's history list.
         """
         if not (300 <= int(self.response.status.split()[0]) < 400):
-            raise AssertionError(
+            raise NotARedirectError(
                 "Can't follow non-redirect response (got %s for %s %s)" % (
                     self.response.status,
                     self.request.request_method,
@@ -1103,9 +1164,24 @@ class TestAgent(object):
         while True:
             try:
                 agent = agent.follow()
-            except AssertionError:
+            except NotARedirectError:
                 return agent
 
+    def new_session(self):
+        """
+        Return a new TestAgent with all cookies deleted.
+        This gives an easy way to test session expiry.
+        """
+        return self.__class__(
+            self.app,
+            self.request,
+            self.response,
+            None,
+            self.history,
+            validate_wsgi=self.validate_wsgi,
+            host=self.host,
+            port=self.port,
+        )
 
     def back(self, count=1):
         return self.history[-abs(count)]
@@ -1141,7 +1217,10 @@ class TestAgent(object):
         if open_in_browser:
             webbrowser.open_new_tab(self.request.make_uri(netloc='{0}:{1}'.format(host, port)))
         print "\nStarting HTTP server on {0}:{1}. Press ctrl-c to exit.".format(host, port)
-        server.serve_forever()
+        try:
+            server.serve_forever()
+        except KeyboardInterrupt:
+            pass
 
     def __enter__(self):
         """
